/*
 * GKlib.h
 * 
 * George's library of most frequently used routines
 *
 * $Id: GKlib.h,v 1.41 2004/11/26 18:58:08 karypis Exp $
 *
 */

#ifndef _GKLIB_H_
#define _GKLIB_H_


/*************************************************************************
* Header file inclusion section
**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <ctype.h>
#include <math.h>
#include <stdarg.h>
#include <time.h>
#include <string.h>
#include <limits.h>
#ifndef _MSC_VER
#include <getopt.h>
#include <sys/time.h>
#include <sys/resource.h>
#endif
#include "getopt.h"




/*************************************************************************
* Data structure definition section
**************************************************************************/
/*-------------------------------------------------------------
 * The following data structure stores int key-value pairs
 *-------------------------------------------------------------*/
struct KeyValueType {
  int key;
  int val;
};

typedef struct KeyValueType KeyValueType;


/*----------------------------------------------------------------
 * The following data structure stores int-key - float-value pairs
 *----------------------------------------------------------------*/
struct FKeyValueType {
  float key;
  int val;
};

typedef struct FKeyValueType FKeyValueType;
/*********************************
 *The following structure was added by Huzefa to store another variable other than key and val
 *-------------------------------------------*/

struct FKeyValueType2 {
    float key;
    int val;
    int extra_var; // will store kmer

};

typedef struct FKeyValueType2 FKeyValueType2;



/*----------------------------------------------------------------
 * The following data structure stores int-key - double-value pairs
 *----------------------------------------------------------------*/
struct DKeyValueType {
  double key;
  int val;
};

typedef struct DKeyValueType DKeyValueType;


/*-------------------------------------------------------------
 * The following data structure stores a sparse CSR format
 *-------------------------------------------------------------*/
struct csrdef {
  int nrows, ncols;
  int *rowptr, *colptr;
  int *rowind, *colind;
  int *irowval, *icolval;
  float *frowval, *fcolval;
};

typedef struct csrdef CSRType;


/*-------------------------------------------------------------
 * The following data structure implements max-heap priority
 * queue
 *-------------------------------------------------------------*/
struct pqueuedef {
  int nnodes;
  int maxnodes;

  /* Heap version of the data structure */
  FKeyValueType *heap;
  int *locator;
};

typedef struct pqueuedef PQueueType;



/*-------------------------------------------------------------
* The following data structure implements a string-2-int mapping
* table used for parsing command-line options
*-------------------------------------------------------------*/
typedef struct {
  char *name;
  int id;
} StringMapType;


/*------------------------------------------------------------
 * This structure implements a simple hash table
 *------------------------------------------------------------*/
typedef struct {
  int nelements;        /* The overall size of the hash-table */
  int htsize;           /* The current size of the hash-table */
  KeyValueType *harray; /* The actual hash-table */
} HTableType;


/*------------------------------------------------------------
 * This structure implements a TokensType list returned by the
 * string tokenizer
 *------------------------------------------------------------*/
typedef struct {
  int ntoks;        /* The number of tokens in the input string */
  char *strbuf;     /* The memory that stores all the entries */
  char **list;      /* Pointers to the strbuf for each element */
} TokensType;



/*************************************************************************
* Definition Section
**************************************************************************/
#define LTERM                   (void **) 0     /* List terminator for GKfree() */

#define HTABLE_EMPTY            -1
#define HTABLE_DELETED          -2
#define HTABLE_FIRST             1
#define HTABLE_NEXT              2



/*************************************************************************
* Macros Section
**************************************************************************/
/*-------------------------------------------------------------
 * Usefull commands 
 *-------------------------------------------------------------*/
#define amax(a, b) ((a) >= (b) ? (a) : (b))
#define amin(a, b) ((a) >= (b) ? (b) : (a))
#define amax3(a, b, c) ((a) >= (b) && (a) >= (c) ? (a) : ((b) >= (a) && (b) >= (c) ? (b) : (c)))
#ifndef _MSC_VER
#define RandomInRange(u)     ((random())%(u))
#else
#define RandomInRange(u) ((int)(1.0*(u)*rand()/(RAND_MAX+1.0)))
#endif
#define SWAP(a, b, tmp) do {(tmp) = (a); (a) = (b); (b) = (tmp);} while(0) 
#define INC_DEC(a, b, val) do {(a) += (val); (b) -= (val);} while(0)
#define icopy(n, a, b) (int *)memmove((void *)(b), (void *)(a), sizeof(int)*(n)) 
#define scopy(n, a, b) (float *)memmove((void *)(b), (void *)(a), sizeof(float)*(n))
#define fcopy(n, a, b) (float *)memmove((void *)(b), (void *)(a), sizeof(float)*(n))
#define dcopy(n, a, b) (double *)memmove((void *)(b), (void *)(a), sizeof(double)*(n))
#define sign(a, b) ((a >= 0 ? b : -b))


/*-------------------------------------------------------------
 * Timing macros
 *-------------------------------------------------------------*/
#define cleartimer(tmr) (tmr = 0.0)
#define starttimer(tmr) (tmr -= seconds())
#define stoptimer(tmr) (tmr += seconds())
#define gettimer(tmr) (tmr)

/*-------------------------------------------------------------
 * dbglvl handling macros
 *-------------------------------------------------------------*/
#define IFSET(a, flag, cmd) if ((a)&(flag)) (cmd);

 
/*-------------------------------------------------------------
 * Debuging memory leaks
 *-------------------------------------------------------------*/
#ifdef DMALLOC
#define imalloc(n, msg) (malloc(sizeof(int)*(n)))
#define fmalloc(n, msg) (malloc(sizeof(float)*(n)))
#define keymalloc(n, msg) (malloc(sizeof(KeyValueType)*(n)))
#define fkeymalloc(n, msg) (malloc(sizeof(FKeyValueType)*(n)))
#define ismalloc(n, val, msg) (iset((n), (val), malloc(sizeof(int)*(n))))
#define fsmalloc(n, val, msg) (fset((n), (val), malloc(sizeof(float)*(n))))
#define dsmalloc(n, val, msg) (dset((n), (val), malloc(sizeof(double)*(n))))
#define GKmalloc(a, b) (malloc((a)))
#endif

#ifdef DMALLOC
#   define MALLOC_CHECK(ptr)                                          \
    if (malloc_verify((ptr)) == DMALLOC_VERIFY_ERROR) {  \
        printf("***MALLOC_CHECK failed on line %d of file %s: " #ptr "\n", \
              __LINE__, __FILE__);                               \
        abort();                                                \
    }
#else
#   define MALLOC_CHECK(ptr) ;
#endif 


/*-------------------------------------------------------------
 * CSR conversion macros
 *-------------------------------------------------------------*/
#define MAKECSR(i, n, a) \
   do { \
     for (i=1; i<n; i++) a[i] += a[i-1]; \
     for (i=n; i>0; i--) a[i] = a[i-1]; \
     a[0] = 0; \
   } while(0) 

#define SHIFTCSR(i, n, a) \
   do { \
     for (i=n; i>0; i--) a[i] = a[i-1]; \
     a[0] = 0; \
   } while(0) 




/*-------------------------------------------------------------
 * Program Assertions
 *-------------------------------------------------------------*/
#ifdef DEBUG
#   define ASSERT(expr)                                          \
    if (!(expr)) {                                               \
        printf("***ASSERTION failed on line %d of file %s: " #expr "\n", \
              __LINE__, __FILE__);                               \
        abort();                                                \
    }
#else
#   define ASSERT(expr) ;
#endif 

#ifdef DEBUG
#   define ASSERTP(expr,msg)                                          \
    if (!(expr)) {                                               \
        printf("***ASSERTION failed on line %d of file %s: " #expr "\n", \
              __LINE__, __FILE__);                               \
        printf msg ; \
        printf("\n"); \
        abort();                                                \
    }
#else
#   define ASSERTP(expr,msg) ;
#endif 




/*************************************************************************
* Function prototypes
**************************************************************************/
/*-------------------------------------------------------------
 * blas.c
 *-------------------------------------------------------------*/
int *iset(int n, int val, int *x);
float *fset(int n, float val, float *x);
double *dset(int n, double val, double *x);
int *iincset(int n, int baseval, int *x);
int iamax(int, int *);
int iamin(int, int *);
int famax(int, float *);
int famin(int, float *);
int damax(int, double *);
int damin(int, double *);
int charsum(int, char *);
int isum(int, int *);
float ssum(int n, float *x, int);
double dsum(int n, double *x, int);
void sscale(int n, float, float *x);
float snorm2(int, float *);
float sdot(int n, float *, float *);
void saxpy(int, float, float *, int, float *, int);
void dscale(int n, double, double *x);
double dnorm2(int, double *);
double ddot(int n, double *, double *);
void daxpy(int, double, double *, int, double *, int);


/*-------------------------------------------------------------
 * file.c
 *-------------------------------------------------------------*/
FILE *GKfopen(char *, char *, char *);
void GKfclose(FILE *);
int GKfexist(char *);
void GKGetFileStats(char *fname, int *r_nlines, int *r_ntokens, int *r_nbytes);

/*-------------------------------------------------------------
 * memory.c
 *-------------------------------------------------------------*/
#ifndef DMALLOC
int *imalloc(int, char *);
int *ismalloc(int, int, char *);
float *fmalloc(int, char *);
float *fsmalloc(int, float, char *);
double *dmalloc(int, char *);
double *dsmalloc(int, double, char *);
KeyValueType *keymalloc(int, char *);
FKeyValueType *fkeymalloc(int, char *);
FKeyValueType2 *fkey2malloc(int, char *); // HR added this
DKeyValueType *dkeymalloc(int, char *);
void *GKmalloc(int, char *);
void GKAllocMatrix(void ***, int, int , int);
void GKFreeMatrix(void **, int, int);
int **iAllocMatrix(int, int , int, char *);
void iFreeMatrix(int ***, int, int);
float **fAllocMatrix(int, int , float, char *);
void fFreeMatrix(float ***, int, int);
char **cAllocMatrix(int, int , char, char *);
void cFreeMatrix(char ***, int, int);
void cSetMatrix(char **, int, int , char);
#endif
void GKfree(void **ptr1,...);
int GKResetMemoryUsage();
int GKGetMaxMemoryUsed();



/*-------------------------------------------------------------
 * util.c
 *-------------------------------------------------------------*/
void *errexit(char *,...);
void RandomPermute(int, int *, int);
void RandomPermuteCoarse(int, int *, int);
void RandomPermuteFine(int, int *, int);
int mylog2(int);
int ispow2(int);
float flog2(float);
double seconds(void);

/*-------------------------------------------------------------
 * string.c
 *-------------------------------------------------------------*/
char *strtprune(char *, char *);
char *strtoupper(char *); 
char *strtolower(char *); 
char *mystrdup(char *orgstr);
int mystrcasecmp(char *s1, char *s2);
int GetStringID(StringMapType *strmap, char *key);


/*-------------------------------------------------------------
 * Sorting routines
 *-------------------------------------------------------------*/
void iintsort(int, int *);
void ifloatsort(int, float *);
void ikeysort(int, KeyValueType *);
void dkeysort(int, KeyValueType *);
void ifkeysort(int, FKeyValueType *);
void dfkeysort(int, FKeyValueType *);
void dfkey2sort(int, FKeyValueType2 *); // HR added this
void idkeysort(int, DKeyValueType *);
int dfkeykselect(int, int, FKeyValueType *);

/*-------------------------------------------------------------
 * Priority queue 
 *-------------------------------------------------------------*/
void PQueueInit(PQueueType *, int);
void PQueueReset(PQueueType *);
void PQueueFree(PQueueType *);
int PQueueGetSize(PQueueType *);
int PQueueInsert(PQueueType *, int, float);
int PQueueDelete(PQueueType *, int);
int PQueueUpdate(PQueueType *, int, float);
int PQueueGetMax(PQueueType *);
int PQueueSeeMax(PQueueType *);
int PQueueLength(PQueueType *);
float PQueueGetKey(PQueueType *);
int PQueueSeeConstraintMax(PQueueType *queue, float maxwgt, double *wgts);
int CheckHeap(PQueueType *);

/*-------------------------------------------------------------
 * HTable routines
 *-------------------------------------------------------------*/
HTableType *HTable_Create(int nelements);
void HTable_Reset(HTableType *htable);
void HTable_Resize(HTableType *htable, int nelements);
void HTable_Insert(HTableType *htable, int key, int val);
void HTable_Delete(HTableType *htable, int key);
int HTable_Search(HTableType *htable, int key);
int HTable_GetNext(HTableType *htable, int key, int *val, int type);
int HTable_SearchAndDelete(HTableType *htable, int key);
void HTable_Destroy(HTableType *htable);
int HTable_HFunction(int nelements, int key);
 

/*-------------------------------------------------------------
 * Tokenizer routines
 *-------------------------------------------------------------*/
void strtokenize(char *line, char *delim, TokensType *tokens);
void freetokens(TokensType *tokens);


#endif
