/*
 * cmdline.c
 *
 * This file parses the command line arguments
 *
 * Started 11/24/01
 * George
 *
 *
 */


#include <aasim.h>


/*-------------------------------------------------------------------
 * Command-line options 
 *-------------------------------------------------------------------*/
static struct option long_options[] = {
  {"simtype",           1,      0,      CMD_SIMTYPE},
  {"wmer",              1,      0,      CMD_KMERLEN},
  {"gocost",            1,      0,      CMD_GOCOST},
  {"gecost",            1,      0,      CMD_GECOST},
  {"shift",             1,      0,      CMD_SHIFT},
  {"help",              0,      0,      CMD_HELP},
  {0,                   0,      0,      0}
};



/*-------------------------------------------------------------------
 * Mappings for the various parameter values
 *-------------------------------------------------------------------*/
static StringMapType simtype_options[] = {
  {"kmer-pssm",          SIMTYPE_KMERPSSM},
  {"sw-pssm",          	SIMTYPE_SWPSSM},
  {"bkmer-pssm",       	SIMTYPE_BKMERPSSM},
  {"abkmer-pssm",       	SIMTYPE_ABKMERPSSM},
  {NULL,                 0}
};



/*-------------------------------------------------------------------
 * Mini help
 *-------------------------------------------------------------------*/
static char helpstr[][100] =
{
" ",
"Usage: aasim   [options] <infile>",
" ",
" Required parameters",
"    infile      Stores the filenames of the sequences in the database.",
" ",
" Optional parameters",
"  -simtype=string",
"     Specifies the method to be used for computing similarity between",
"     the sequences.",
"     The possible values are:",
"        kmer-pssm    - A kmer model using the PSSM matrix",
"        sw-pssm      - A SW-based model using a PSSM scoring matrix",
"        bkmer-pssm   - A best-kmer per position based kernel",
"        abkmer-pssm  - A best-kmer per position based kernel using a kmer range.",
" ",
" ",
"  -wmer=int",
"     Specifies the length of the wmer. The default value is 2.",
" ",
"  -gocost=float",
"     Specifies the gap openning cost. The default value is 10.",
" ",
"  -gecost=float",
"     Specifies the gap extension cost. The default value is 2.",
" ",
"  -shift=float",
"     Specifies the value to be added to the p2pscore. Default is 0.0.",
" ",
"  -help",
"     Prints this message.",
""
};

static char shorthelpstr[][100] = {
" ",
"   Usage: aasim [options] <infile>",
"          use 'aasim -help' for a summary of the options."
};
 


/*************************************************************************
* This is the entry point of the command-line argument parser
**************************************************************************/
void parse_cmdline_aasim(CtrlType *ctrl, int argc, char *argv[])
{  
  int i, j, k;
  int c, option_index;

  /* initialize the ctrl data structure */
  ctrl->SimType     = SIMTYPE_KMERPSSM;
  ctrl->SMType      = SMTYPE_BLOSUM62;
  ctrl->kmerLen     = 2;
  ctrl->GOCost      = 10;
  ctrl->GECost      = 2;
  ctrl->shift       = 0.0;
  ctrl->infile      = NULL;


  /* Parse the command line arguments  */
  while ((c = getopt_long_only(argc, argv, "", long_options, &option_index)) != -1) {
      
    switch (c) {
      case CMD_SIMTYPE:
        if (optarg)
          if ((ctrl->SimType = GetStringID(simtype_options, optarg)) == -1)
            errexit("Invalid option -%s=%s\n", long_options[option_index].name, optarg);
        break;
      case CMD_KMERLEN:
        if (optarg) ctrl->kmerLen = atoi(optarg);
        break;
    
      case CMD_GOCOST:
        if (optarg) ctrl->GOCost = atof(optarg);
        break;
      case CMD_GECOST:
        if (optarg) ctrl->GECost = atof(optarg);
        break;
    case CMD_SHIFT:
        if (optarg) ctrl->shift = atof(optarg);
        break;
	
    case CMD_HELP:
        for (i=0; strlen(helpstr[i]) > 0; i++)
          printf("%s\n", helpstr[i]);
        exit(0);
        break;
      case '?':
      default:
        printf("Illegal command-line option(s)\nUse %s -help for a summary of the options.\n", argv[0]);
        exit(0);
    }
  }

  if (argc-optind != 1) {
    printf("Missing parameters.");
    for (i=0; strlen(shorthelpstr[i]) > 0; i++)
      printf("%s\n", shorthelpstr[i]);
    exit(0);
  }

  ctrl->infile = strdup(argv[optind++]);

}




