/* Version 1.2 */
/* Release Date: 03-28-06 */
/* Authors: */
/* George Karypis  (karypis@cs.umn.edu)	 */
/* Huzefa Rangwala (rangwala@cs.umn.edu)      */

/* LICENSE: Please read LICENSE.txt */

/* Please Cite: */
/*      Huzefa Rangwala and George Karypis,	 */
/*      Profile-based direct kernels for remote homology detection and fold recognition.	      */
/*      Bioinformatics 2005 21(23):4239-4247 */


/*
 * io.c 
 *
 * This file contains various I/O routines
 *
 * Started 11/23/04
 * George
 * 
 */

#include <aasim.h>



/******************************************************************************
* Read the file containing the various sequence filenames
*******************************************************************************/
void ReadSequences(CtrlType *ctrl, SeqDBType *seqDB)
{
  int i, nseqs, ntokens, nbytes;
  FILE *fpin;
  TokensType tokens;
  char line[MAXLINELEN];

  GKGetFileStats(ctrl->infile, &nseqs, &ntokens, &nbytes);
  
  InitSeqDB(seqDB);

  seqDB->nseqs     = nseqs;
  seqDB->sequences = (SeqType *)GKmalloc(nseqs*sizeof(SeqType), "ReadSequences: seqDB->sequences");
  seqDB->family    = (char **)GKmalloc(nseqs*sizeof(char *), "ReadSequences: seqDB->family");

  LocalizeSequences(seqDB, ctrl->aaorder);
  
  
  /* Read the sequences one by one */
  fpin = GKfopen(ctrl->infile, "r", "ReadSequences: fpin");
  for (i=0; i<nseqs; i++) {
    fgets(line, MAXLINELEN-1, fpin); 
    strtokenize(line, " \t\n", &tokens);


    ReadOneSequence_PSIBLAST(ctrl, seqDB->sequences+i, tokens.list[0], seqDB->i2c, seqDB->c2i);

    seqDB->family[i] = (tokens.ntoks > 1 ? strdup(tokens.list[1]) : strdup("Fam1"));
    strtoupper(seqDB->family[i]);

    freetokenslist(&tokens);
  }
  GKfclose(fpin);

  SetUpClasses(ctrl, seqDB);

}




/******************************************************************************
* Reads in a file storing a PSSM+PSFM in GK's simplified PSI-Blast format
*******************************************************************************/
void ReadOneSequence_PSIBLAST(CtrlType *ctrl, SeqType *seq, char *filename, int *i2c, int *c2i)
{
  int i, j, k, len, ntokens, nbytes, printseq=0,count;
  FILE *fpin;
  char header[PSSMWIDTH], line[MAXLINELEN];
  TokensType tokens;


  GKGetFileStats(filename, &len, &ntokens, &nbytes);
  len--; /* skip the header row */
  
  seq->len      = len;
  seq->name     = strdup(filename);
  seq->sequence = imalloc(len, "ReadOneSequence_PSIBLAST: sequence");
  seq->pssm = iAllocMatrix(len, PSSMWIDTH, 0, "ReadOneSequence_PSIBLAST:seq->pssm");
  seq->psfm = iAllocMatrix(len, PSSMWIDTH, 0, "ReadOneSequence_PSIBLAST:seq->psfm");

  
  fpin = GKfopen(filename, "r", "ReadOneSequence_PSIBLAST: fpin");


  /* Read the header line */
  fgets(line, MAXLINELEN-1, fpin); 
  strtoupper(line);
  strtokenize(line, " \t\n", &tokens);

  for (i=0; i<PSSMWIDTH; i++)
    header[i] = tokens.list[i][0];

  freetokenslist(&tokens);


  /* Read the rest of the lines */
  for (i=0,count=0;count<len;count++){
    fgets(line, MAXLINELEN-1, fpin); 
    strtoupper(line);
    strtokenize(line, " \t\n", &tokens);
    
    if(tokens.list[1][0] != 'X' && tokens.list[1][0] != 'B'){
	seq->sequence[i] = c2i[tokens.list[1][0]];

	for (j=0; j<PSSMWIDTH; j++) 
	    seq->pssm[i][c2i[header[j]]] = atoi(tokens.list[2+j]);
	
	for (j=0; j<PSSMWIDTH; j++) 
	    seq->psfm[i][c2i[header[j]]] = atoi(tokens.list[2+PSSMWIDTH+j]);
	
	/* take care the instances in which PSFM is empty */
	if (isum(PSSMWIDTH, seq->psfm[i]) == 0) {
	    //      printseq = 1;
	    //      printf("Sequence %s does not have a PSFM at %d...\n", filename, i); 
	    
	    for (j=0; j<PSSMWIDTH; j++)
        seq->psfm[i][j] = ctrl->aafmat[seq->sequence[i]][j];
	    
	    /* The following does not work as it leads to exremenly biased scores!
	       iset(PSSMWIDTH, 1, seq->psfm[i]);
	       seq->psfm[i][seq->sequence[i]] = 81;
*/
	}
	i ++;
    }

	freetokenslist(&tokens);
  }

  GKfclose(fpin);


}




/******************************************************************************
* This function writes the matrix in the format suitable for cluto
*******************************************************************************/
void WriteSimMatrix(CtrlType *ctrl, SeqDBType *seqDB, MatType *mat)
{
  int i, j, k;
  char filename[1024];
  FILE *fpout;
  
  //  printf("Write the matrix file\n");
  sprintf(filename, "%s.smat",   ctrl->infile);
  
  fpout = GKfopen(filename, "w", "WriteSimMatrix: ds.mat");
  //  fprintf(fpout, "%d %d\n", mat->nrows, mat->rowptr[mat->nrows]);
  for (i=0; i<mat->nrows; i++) {
    for (j=mat->rowptr[i]; j<mat->rowptr[i+1] -1; j++) 
	fprintf(fpout,"%f ", mat->rowval[j]);
    fprintf(fpout, "%f\n",mat->rowval[mat->rowptr[i+1]-1]);
  }
  GKfclose(fpout);

  strcat(filename, ".rclass");
  fpout = GKfopen(filename, "w", "WriteMatrix: filename.rclass");
  for (i=0; i<mat->nrows; i++) {
    fprintf(fpout, "%d\n", seqDB->class[i]);
  }
  GKfclose(fpout);

}
