/* Version 1.2 */
/* Release Date: 03-28-06 */
/* Authors: */
/* George Karypis  (karypis@cs.umn.edu)	 */
/* Huzefa Rangwala (rangwala@cs.umn.edu)      */

/* LICENSE: Please read LICENSE.txt */

/* Please Cite: */
/*      Huzefa Rangwala and George Karypis,	 */
/*      Profile-based direct kernels for remote homology detection and fold recognition.	      */
/*      Bioinformatics 2005 21(23):4239-4247 */


// Modified by Huzefa Rangwala
/*
 * sim.c
 *
 * This file contains the various routines for computing the seq-2-seq similarities
 *
 * Started 11/26/04
 */


#include <aasim.h>


/***************************************************************************
* This is the entry point of the seq-2-seq similarity calculations routines
****************************************************************************/
void ComputeSeq2SeqSimilarities(CtrlType *ctrl, SeqDBType *seqDB, MatType *smat)
{

  switch (ctrl->SimType) {
  case SIMTYPE_KMERPSSM:
      ComputeSimilaritiesKMERPSSM_FAST(ctrl, seqDB, smat);
      break;
  case SIMTYPE_SWPSSM:
      ComputeSimilaritiesSWPSSM(ctrl, seqDB, smat);
      break;
  case SIMTYPE_BKMERPSSM:
      ComputeSimilaritiesBKMERPSSM(ctrl, seqDB, smat);
      break;
  case SIMTYPE_ABKMERPSSM:
      ComputeSimilaritiesABKMERPSSM2(ctrl, seqDB, smat);
      break;
  default:
      errexit("Unknown SimType of %d\n", ctrl->SimType);
  }

}





/***************************************************************************
* This function computes k-mer based similarities using a global amino acid
* scoring matrix.
****************************************************************************/
void ComputeSimilaritiesKMERPSSM_FAST(CtrlType *ctrl, SeqDBType *seqDB, MatType *smat)
{
  int i, j, ii, jj, k, diag, nseqs, wlen, ilen, jlen, dpsize, maxseqlen;
  int *iseq, *jseq, **ipssm, **jpssm, **ipsfm, **jpsfm;
  float **p2pscore, **sims, sim, score;
  char **dotplot;

  nseqs  = seqDB->nseqs;
  wlen   = 2 * ctrl->kmerLen + 1;

  /* Allocate memory for a dense matrix for now */
  sims     = fAllocMatrix(nseqs, nseqs, 0.0, "ComputeSimilaritiesKMERPSSM: sims");

  /* Determine the maximum sequence length and allocate memory for the pairwise position scores */
  for (maxseqlen=0, i=0; i<nseqs; i++) 
    maxseqlen = (seqDB->sequences[i].len > maxseqlen ? seqDB->sequences[i].len : maxseqlen);
  p2pscore = fAllocMatrix(maxseqlen, maxseqlen, 0.0, "ComputeSimilaritiesKMERPSSM: p2pscore");


  
  /* Get into the main loop */ 
  for (i=0; i<nseqs; i++) {
    ilen  = seqDB->sequences[i].len;
    iseq  = seqDB->sequences[i].sequence;
    ipssm = seqDB->sequences[i].pssm;
    ipsfm = seqDB->sequences[i].psfm;

    for (j=i; j<nseqs; j++) {
      jlen  = seqDB->sequences[j].len;
      jseq  = seqDB->sequences[j].sequence;
      jpssm = seqDB->sequences[j].pssm;
      jpsfm = seqDB->sequences[j].psfm;

      
      /* Compute the similarity between each individual pair of columns */
      ComputeP2PScore(ctrl, p2pscore, ilen, ipssm, ipsfm, jlen, jpssm, jpsfm);
      

      /* Compute the actual similarity */
      sim = 0.0;
      if (ilen >= wlen && jlen >= wlen) {
	  /* take care positive diagonals */
        for (diag=-(ilen-wlen); diag<=(jlen-wlen); diag++) {
          ii = (diag <  0  ? -diag : 0);
          jj = (diag >= 0  ?  diag : 0);

          /* take care the initial score */
          for (score=0.0, k=0; k<wlen; k++) 
            score += p2pscore[ii+k][jj+k];

	  if(score > 0.0)
	      sim += score;

          /* incrementally compute the other scores along the diagonal */
          while (ii<ilen-wlen && jj<jlen-wlen) {
            score = score - p2pscore[ii][jj] + p2pscore[ii+wlen][jj+wlen];
            ii++; jj++;

	    if(score > 0.0)
		sim += score;
            
          }
        }
      }
      sims[i][j] = sims[j][i] = sim;

    }
  }
  
  
  
  CreateSimMatrix(ctrl, seqDB, smat, sims);

  fFreeMatrix(&sims, nseqs, nseqs);
  fFreeMatrix(&p2pscore, maxseqlen, maxseqlen);
  
  
}




/***************************************************************************
* This function computes the similarity between two sequences based
* on the SW-score using a PSSM amino acid scoring matrix.
****************************************************************************/
void ComputeSimilaritiesSWPSSM(CtrlType *ctrl, SeqDBType *seqDB, MatType *smat)
{
  int i, j, ii, jj, k, nseqs, ilen, jlen, maxseqlen, dpsize, lalen;
  int *iseq, *jseq, **ipssm, **ipsfm, **jpssm, **jpsfm, *ila, *jla;
  float **p2pscore, **sims, sim;
  char **dotplot;

  nseqs  = seqDB->nseqs;


  /* Allocate memory for a dense matrix for now */
  sims = fAllocMatrix(nseqs, nseqs, 0.0, "ComputeSimilaritiesSWGSM: sims");

  /* Determine the maximum sequence length and allocate memory for the local alignment arrays */
  for (maxseqlen=0, i=0; i<nseqs; i++) 
    maxseqlen = (seqDB->sequences[i].len > maxseqlen ? seqDB->sequences[i].len : maxseqlen);
  ila = imalloc(2*maxseqlen, "ComputeSimilaritiesSWGSM: ila");
  jla = imalloc(2*maxseqlen, "ComputeSimilaritiesSWGSM: jla");

  p2pscore = fAllocMatrix(maxseqlen, maxseqlen, 0.0, "ComputeSimilaritiesSWPSSM: p2pscore");

  


  /* Get into the main loop for computing the similarities */ 
  for (i=0; i<nseqs; i++) {
    ilen  = seqDB->sequences[i].len;
    iseq  = seqDB->sequences[i].sequence;
    ipssm = seqDB->sequences[i].pssm;
    ipsfm = seqDB->sequences[i].psfm;

    for (j=i; j<nseqs; j++) {
      jlen  = seqDB->sequences[j].len;
      jseq  = seqDB->sequences[j].sequence;
      jpssm = seqDB->sequences[j].pssm;
      jpsfm = seqDB->sequences[j].psfm;

      
      /* Compute the similarity between each individual pair of columns */
      ComputeP2PScore(ctrl, p2pscore, ilen, ipssm, ipsfm, jlen, jpssm, jpsfm);
      
      sim = SWAlign_PSSM(ilen, jlen, p2pscore, ctrl->GOCost, ctrl->GECost, &lalen, ila, jla);
      
      

      sims[i][j] = sims[j][i] = sim;

    }
  }
  

  
  CreateSimMatrix(ctrl, seqDB, smat, sims);

  GKfree((void *)&ila, &jla, LTERM);
  
  fFreeMatrix(&sims, nseqs, nseqs);
  fFreeMatrix(&p2pscore, maxseqlen, maxseqlen);

  
}


/***************************************************************************
* This function computes k-mer based similarities using a PSSM matrix. 
.
****************************************************************************/
void ComputeSimilaritiesBKMERPSSM(CtrlType *ctrl, SeqDBType *seqDB, MatType *smat)
{
  int i, j, ii, jj, k, diag, nseqs, wlen, ilen, jlen, dpsize, maxseqlen, ncand;
  int *iseq, *jseq, **ipssm, **jpssm, **ipsfm, **jpsfm, *imarker, *jmarker;
  float **p2pscore, **sims, sim, score;
  FKeyValueType *fcand;
  char **dotplot;

  nseqs  = seqDB->nseqs;
  wlen   = 2 * ctrl->kmerLen + 1;

  /* Allocate memory for a dense matrix for now */
  sims     = fAllocMatrix(nseqs, nseqs, 0.0, "ComputeSimilaritiesKMERPSSM: sims");

  /* Determine the maximum sequence length and allocate memory for the pairwise position scores */
  for (maxseqlen=0, i=0; i<nseqs; i++) 
      maxseqlen = (seqDB->sequences[i].len > maxseqlen ? seqDB->sequences[i].len : maxseqlen);
  
  p2pscore = fAllocMatrix(maxseqlen, maxseqlen, 0.0, "ComputeSimilaritiesKMERPSSM: p2pscore");

  /* Allocate memory for the candidates */
  imarker = imalloc(maxseqlen, "ComputeSimilaritiesBKMERPSSM: imarker");
  jmarker = imalloc(maxseqlen, "ComputeSimilaritiesBKMERPSSM: jmarker");
  fcand   = fkeymalloc(maxseqlen*maxseqlen, "ComputeSimilaritiesBKMERPSSM: fcand");

  
  /* Get into the main loop */ 
  for (i=0; i<nseqs; i++) {
    ilen  = seqDB->sequences[i].len;
    iseq  = seqDB->sequences[i].sequence;
    ipssm = seqDB->sequences[i].pssm;
    ipsfm = seqDB->sequences[i].psfm;

    for (j=i; j<nseqs; j++) {
      jlen  = seqDB->sequences[j].len;
      jseq  = seqDB->sequences[j].sequence;
      jpssm = seqDB->sequences[j].pssm;
      jpsfm = seqDB->sequences[j].psfm;

  
      /* Compute the similarity between each individual pair of columns */
      ComputeP2PScore(ctrl, p2pscore, ilen, ipssm, ipsfm, jlen, jpssm, jpsfm);


      /* Compute the actual similarity */
      ncand = 0;
      if (ilen >= wlen && jlen >= wlen) {
	  /* take care positive diagonals */
	  for (diag=-(ilen-wlen); diag<=(jlen-wlen); diag++) {
          ii = (diag <  0  ? -diag : 0);
          jj = (diag >= 0  ?  diag : 0);

          /* take care the initial score */
          for (score=0.0, k=0; k<wlen; k++) 
            score += p2pscore[ii+k][jj+k];

          if (score > 0) {
            fcand[ncand].val   = ii*maxseqlen + jj;
            fcand[ncand++].key = score;
          }

          /* incrementally compute the other scores along the diagonal */
          while (ii<ilen-wlen && jj<jlen-wlen) {
            score = score - p2pscore[ii][jj] + p2pscore[ii+wlen][jj+wlen];
            ii++; jj++;

            if (score > 0) {
              fcand[ncand].val   = ii*maxseqlen + jj;
              fcand[ncand++].key = score;
            }
          }
        }
      }


      /* Select the best kmer pairs in a greedy fashion */
      dfkeysort(ncand, fcand);
      iset(ilen, 0, imarker);
      iset(jlen, 0, jmarker);
      for (sim=0.0, k=0; k<ncand; k++) {
        ii = fcand[k].val/maxseqlen;
        jj = fcand[k].val%maxseqlen;
        if (imarker[ii] < 1 && jmarker[jj] < 1) {
          sim += fcand[k].key;
          imarker[ii]++;
          jmarker[jj]++;

        }
      }

      sims[i][j] = sims[j][i] = sim;

      
    }
  }
  

  
  CreateSimMatrix(ctrl, seqDB, smat, sims);

  GKfree((void *)&imarker, &jmarker, &fcand, LTERM);
  fFreeMatrix(&sims, nseqs, nseqs);
  fFreeMatrix(&p2pscore, maxseqlen, maxseqlen);

  
}




/***************************************************************************
* This function computes all k-mer based similarities using a PSSM.
****************************************************************************/
void ComputeSimilaritiesABKMERPSSM2(CtrlType *ctrl, SeqDBType *seqDB, MatType *smat)
{
  int i, j, ii, jj, k, nseqs, wlen, minwlen, ilen, jlen, dpsize, maxseqlen, ncand;
  int *iseq, *jseq, *imarker, *jmarker;
  float **aasmat, **p2pscore, **sims, sim, *scores, r1, r2;
  FKeyValueType *fcand;
  char **dotplot;

  nseqs   = seqDB->nseqs;
  wlen    = ctrl->kmerLen;
  minwlen = 0;
  aasmat  = ctrl->aasmat;

  /* Determine the maximum sequence length and allocate memory for the pairwise position scores */
  for (maxseqlen=0, i=0; i<nseqs; i++) 
    maxseqlen = (seqDB->sequences[i].len > maxseqlen ? seqDB->sequences[i].len : maxseqlen);
  p2pscore = fAllocMatrix(maxseqlen, maxseqlen, 0.0, "ComputeSimilaritiesKMERPSSM: p2pscore");

  /* Allocate memory for the candidates */
  imarker = imalloc(maxseqlen, "ComputeSimilaritiesBKMERPSSM: imarker");
  jmarker = imalloc(maxseqlen, "ComputeSimilaritiesBKMERPSSM: jmarker");
  fcand   = fkeymalloc(maxseqlen*maxseqlen, "ComputeSimilaritiesBKMERPSSM: fcand");

  /* Allocate memory for a dense matrix for now */
  sims = fAllocMatrix(nseqs, nseqs, 0.0, "ComputeSimilaritiesKMER: sims");

  scores  = fmalloc(wlen+1, "ComputeSimilaritiesABKMERGSM: scores");

  
  /* Get into the main loop for computing the similarities */ 
  for (i=0; i<nseqs; i++) {
    ilen  = seqDB->sequences[i].len;
    iseq  = seqDB->sequences[i].sequence;

    for (j=i; j<nseqs; j++) {
      jlen  = seqDB->sequences[j].len;
      jseq  = seqDB->sequences[j].sequence;

  
      /* Compute the similarity between each individual pair of columns */
      ComputeP2PScore(ctrl, p2pscore, ilen, seqDB->sequences[i].pssm, seqDB->sequences[i].psfm, 
                                jlen, seqDB->sequences[j].pssm, seqDB->sequences[j].psfm);

      /* Compute the actual similarity */
      ncand = 0;
      for (ii=wlen; ii<ilen-wlen; ii++) {
        for (jj=wlen; jj<jlen-wlen; jj++) {
          scores[0] = p2pscore[ii][jj];
          for (k=1; k<=wlen; k++) 
            scores[k]  = scores[k-1] + p2pscore[ii-k][jj-k] + p2pscore[ii+k][jj+k];

          /* find the maximum score */
          for (k=wlen-1; k>=minwlen; k--) 
            scores[wlen] = amax(scores[wlen], scores[k]);

          if (scores[wlen] > 0) {
            fcand[ncand].val   = ii*maxseqlen + jj;
            fcand[ncand++].key = scores[wlen];
          }
        }
      }

#ifdef XXX
      for (ii=0; ii<ilen; ii++) {
        for (jj=0; jj<jlen; jj++) {
          scores[0] = p2pscore[ii][jj];
          for (k=1; k<=wlen; k++) {
            r1 = (ii-k < 0 || jj-k < 0 ? 0.0 : p2pscore[ii-k][jj-k]);
            r2 = (ii+k >= ilen || jj+k >= jlen ? 0.0 : p2pscore[ii+k][jj+k]);
            scores[k]  = scores[k-1] + r1 + r2;
          }

          /* find the maximum score */
          for (k=wlen-1; k>=minwlen; k--) 
            scores[wlen] = amax(scores[wlen], scores[k]);

          if (scores[wlen] > 0) {
            fcand[ncand].val   = ii*maxseqlen + jj;
            fcand[ncand++].key = scores[wlen];
          }
        }
      }
#endif

      /* Select the best kmer pairs in a greedy fashion */
      dfkeysort(ncand, fcand);
      iset(ilen, 0, imarker);
      iset(jlen, 0, jmarker);
      for (sim=0.0, k=0; k<ncand; k++) {
        ii = fcand[k].val/maxseqlen;
        jj = fcand[k].val%maxseqlen;
        if (imarker[ii] < 1 && jmarker[jj] < 1) {
          sim += fcand[k].key;
          imarker[ii]++;
          jmarker[jj]++;


        }
      }
      sims[i][j] = sims[j][i] = sim;


    }
  }
  
  GKfree((void *)&imarker, &jmarker, &fcand, &scores, LTERM);
  fFreeMatrix(&p2pscore, maxseqlen, maxseqlen);

  
  CreateSimMatrix(ctrl, seqDB, smat, sims);

  fFreeMatrix(&sims, nseqs, nseqs);

  
}





/***************************************************************************
* This file creates a MatType datastructure our of a sims[][] array
****************************************************************************/
void CreateSimMatrix(CtrlType *ctrl, SeqDBType *seqDB, MatType *smat, float **sims)
{
  int i, j, nseqs, nnz;
  int *rowptr, *rowind;
  float *rowval;

  nseqs = seqDB->nseqs;

  for (nnz=0, i=0; i<nseqs; i++) {
    for (j=0; j<nseqs; j++)
      nnz += (sims[i][j] > 0.0 ? 1 : 0);
  }

  InitMatrix(smat);
  smat->nrows = smat->nrows = nseqs;
  rowptr = smat->rowptr = ismalloc(nseqs+1, 0, "CreateSimMatrix: rowptr");
  rowind = smat->rowind = imalloc(nnz, "CreateSimMatrix: rowind");
  rowval = smat->rowval = fmalloc(nnz, "CreateSimMatrix: rowval");

  
  for (nnz=0, i=0; i<nseqs; i++) {
    for (j=0; j<nseqs; j++) {
      if (sims[i][j] > 0.0) {
        rowind[nnz] = j;
        rowval[nnz] = sims[i][j];
        nnz++;
      }
    }
    rowptr[i+1] = nnz;
  }

}





float SWAlign_PSSM(int ilen, int jlen, float **p2pcost, float gocost, 
                   float gecost, int *r_lalen, int *ila, int *jla)
{
  int i, j, k, lalen, ibest, jbest, clen, state;
  float **V, **E, **F, bestscore;

  V = fAllocMatrix(ilen+1, jlen+1, 0.0, "SWAlign_GSM: V");
  E = fAllocMatrix(ilen+1, jlen+1, 0.0, "SWAlign_GSM: E");
  F = fAllocMatrix(ilen+1, jlen+1, 0.0, "SWAlign_GSM: F");

  /* Compute the score of the alignment */
  bestscore = 0.0;
  ibest = jbest = 0;
  for (i=1; i<=ilen; i++) {
    for (j=1; j<=jlen; j++) {
      F[i][j] = amax(F[i-1][j], V[i-1][j]-gocost) - gecost;
      E[i][j] = amax(E[i][j-1], V[i][j-1]-gocost) - gecost;
      V[i][j] = amax(F[i][j], E[i][j]);
      V[i][j] = amax(V[i-1][j-1]+p2pcost[i-1][j-1], V[i][j]);
      V[i][j] = amax(0.0, V[i][j]);

      if (bestscore < V[i][j]) {
        bestscore = V[i][j];
        ibest = i;
        jbest = j;
      }
    }
  }

  /* Record the alignment. state=1,2,3 for V, F, & E, respectively  */
  for (state=1, clen=0, i=ibest, j=jbest; i>0 && j>0;) {
    if (state == 1) { /* The V[][] case */
      if (V[i][j] == 0.0) {
        break;
      } 
      else if (V[i][j] == F[i][j]) {
        state = 2;
      }
      else if (V[i][j] == E[i][j]) {
        state = 3;
      }
      else {
        ila[clen]   = --i;
        jla[clen++] = --j;
      }
    }
    else if (state == 2) { /* The F[][] case */
      if (F[i][j] == V[i-1][j]-gocost-gecost)
        state = 1;
      ila[clen]   = --i;
      jla[clen++] = -1;
    }
    else if (state == 3) { /* The E[][] case */
      if (E[i][j] == V[i][j-1]-gocost-gecost)
        state = 1;
      ila[clen]   = -1;
      jla[clen++] = --j;
    }
  }

  /* swap the order of the residues in ila and jla */
  for (i=0; i<clen/2; i++) {
    SWAP (ila[i], ila[clen-i-1], j);
    SWAP (jla[i], jla[clen-i-1], j);
  }



  
  fFreeMatrix(&V, ilen+1, jlen+1);
  fFreeMatrix(&E, ilen+1, jlen+1);
  fFreeMatrix(&F, ilen+1, jlen+1);

  *r_lalen = clen;
  return bestscore;
}


/****************************************************************************
* This function computes the position-position score based on the PSSM & PSFM
*****************************************************************************/
void ComputeP2PScore(CtrlType *ctrl, float **p2pscore, int ilen, int **ipssm, 
                     int **ipsfm, int jlen, int **jpssm, int **jpsfm)
{
  int i, j, k, l;
  float i2jscore, j2iscore, sum, localshift;
  float *ipsfmisum, *jpsfmisum, **p2pscore_t;
  

  ipsfmisum = fmalloc(ilen, "ComputeP2PScore: ipsfmisum");
  jpsfmisum = fmalloc(jlen, "ComputeP2PScore: jpsfmisum");
  
  for (i=0; i<ilen; i++) 
        ipsfmisum[i] = 1.0/isum(PSSMWIDTH, ipsfm[i]);
  for (i=0; i<jlen; i++) 
      jpsfmisum[i] = 1.0/isum(PSSMWIDTH, jpsfm[i]);
  
  for (i=0; i<ilen; i++) {
        for (j=0; j<jlen; j++) {
	    i2jscore = j2iscore = 0.0;
	    for (k=0; k<PSSMWIDTH; k++) {
		i2jscore += ipsfm[i][k]*jpssm[j][k];
		j2iscore += jpsfm[j][k]*ipssm[i][k];
          }
	    p2pscore[i][j] = i2jscore*ipsfmisum[i] + j2iscore*jpsfmisum[j];
        }
  }
  
  GKfree((void *)&ipsfmisum, &jpsfmisum, LTERM);

  
  /*---------------------------------------------------------------------
   * shift the scoring system           
   *---------------------------------------------------------------------*/
  localshift = ctrl->shift;
  
  
  for (i=0; i<ilen; i++) {
      for (j=0; j<jlen; j++) {
	  p2pscore[i][j] += localshift;
	  
      /* Keep three decimal points of significance */
	  p2pscore[i][j] = rint(1000*p2pscore[i][j])/1000;  
      }
  }
  
}

