/*
 * util.c 
 *
 * This file contains various I/O routines
 *
 * Started 11/23/04
 * George
 * 
 * $Id: util.c,v 1.5 2004/11/26 17:53:25 karypis Exp $
 *
 */

#include <aasim.h>


/******************************************************************************
* This function localizes the labels of the sequences for efficient processing
*******************************************************************************/
void LocalizeSequences(SeqDBType *sdb, char *alphabet)
{
  int i;

  sdb->nsymbols = strlen(alphabet);

  iset(256, -1, sdb->i2c);
  iset(256, -1, sdb->c2i);

  for (i=0; i<sdb->nsymbols; i++) {
    sdb->i2c[i]           = alphabet[i];
    sdb->c2i[alphabet[i]] = i;
  }

}


/*************************************************************************
* This function allocate memory for a SeqDBType and initializes it
**************************************************************************/
SeqDBType *CreateSeqDB()
{
  SeqDBType *sdb;

  sdb = (SeqDBType *)GKmalloc(sizeof(SeqDBType), "CreateSeqDB: sdb");

  InitSeqDB(sdb);

  return sdb;
}


/*************************************************************************
* This function initializes a sequence DB
**************************************************************************/
void InitSeqDB(SeqDBType *sdb)
{
  sdb->nseqs     = 0;
  sdb->nsymbols  = 0;
  sdb->sequences = NULL;
  sdb->family    = NULL;
}


/*************************************************************************
* This function frees the memory allocated for a sequnce db
**************************************************************************/
void FreeSeqDB(SeqDBType *sdb, int freetop)
{
  int i;

  for (i=0; i<sdb->nseqs; i++) {
    iFreeMatrix(&sdb->sequences[i].pssm, sdb->sequences[i].len, PSSMWIDTH);
    iFreeMatrix(&sdb->sequences[i].psfm, sdb->sequences[i].len, PSSMWIDTH);
    GKfree((void *)&(sdb->sequences[i].sequence), &sdb->sequences[i].name, LTERM);
    GKfree((void *)&(sdb->family[i]), LTERM);
  }

  GKfree((void *)&sdb->sequences, &sdb->family, LTERM); 

  if (freetop)
    free(sdb);
}


/*************************************************************************
* This function allocate memory for a matrix and initializes it
**************************************************************************/
MatType *CreateMatrix()
{
  MatType *mat;

  mat = (MatType *)GKmalloc(sizeof(MatType), "CreateMatrix: mat");

  InitMatrix(mat);

  return mat;
}


/*************************************************************************
* This function initializes the matrix
**************************************************************************/
void InitMatrix(MatType *mat)
{
  mat->nrows = mat->ncols = mat->nclasses = -1;
  mat->rowptr = mat->rowind = NULL;
  mat->rowval = NULL;
  mat->class = NULL;
}


/*************************************************************************
* This function frees the memory allocated for a matrix
**************************************************************************/
void FreeMatrix(MatType *mat, int freetop)
{
  GKfree((void *)&mat->rowptr, &mat->rowind, &mat->rowval, &mat->class, LTERM);

  if (freetop)
    free(mat);
}
                 


/*************************************************************************
* This function creates determines the classes and creates a class vector
**************************************************************************/
void SetUpClasses(CtrlType *ctrl, SeqDBType *seqDB)
{
  int i, j, k, nseqs, nclasses;
  int *class;
  char **family, **classnames;
  char str[256];

  nseqs  = seqDB->nseqs;
  family = seqDB->family;

  class = seqDB->class = imalloc(nseqs, "SetUpClasses: class");

  for (nclasses=0, i=0; i<nseqs; i++) {
    for (j=0; j<i && strcmp(family[i], family[j]); j++);
    class[i] = (j < i ? class[j] : nclasses++);
  }

  seqDB->nclasses = nclasses;

  classnames = seqDB->classnames = (char **)GKmalloc(nclasses*sizeof(char *), "SetUpClasses: classnames");
  for (i=0; i<nclasses; i++)
    classnames[i] = NULL;

  for (i=0; i<nseqs; i++) {
    if (classnames[class[i]] == NULL) 
      classnames[class[i]] = strdup(family[i]);
  }
}


/*****************************************************************************
* This function prints the contents of a SeqType data structure
******************************************************************************/
void PrintSeqType(SeqType *seq, int *i2c, int *c2i)
{
  int i, j, k, len;

  printf("Sequence: %s\n", seq->name);
  printf("      ");
  for (i=0; i<PSSMWIDTH; i++)
    printf("  %c", (char)i2c[i]);
  printf("   ");
  for (i=0; i<PSSMWIDTH; i++)
    printf("  %c", (char)i2c[i]);
  printf("\n");

  for (i=0; i<seq->len; i++) {
    printf("%3d %c ", i, i2c[seq->sequence[i]]);

    for (j=0; j<PSSMWIDTH; j++)
      printf("%3d", seq->pssm[i][j]);
    printf("   ");
    for (j=0; j<PSSMWIDTH; j++)
      printf("%3d", seq->psfm[i][j]);
    printf("\n");
  }
}
  

